//Libraries
#include <Arduino.h>
#include <Wire.h>
#include "Adafruit_SHT31.h"
#include <Adafruit_Sensor.h>
#include "Adafruit_TSL2591.h"
#include "RTClib.h"
#include "Adafruit_BLE.h"
#include "Adafruit_BluefruitLE_SPI.h"
//#include "Adafruit_BluefruitLE_UART.h"
#include "BluefruitConfig.h"
#include <SPI.h>
#include <SD.h>


#if SOFTWARE_SERIAL_AVAILABLE
  #include <SoftwareSerial.h>
#endif

//Define
int OnePassFlag = 0;        //Flag for Startup Run
bool enableHeater = false;  //Temp&Humid Sensor Heater Flag
uint8_t loopCnt = 0;
char daysOfTheWeek[7][12] = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"}; //RTC Variable
char dateTimeString[30]; // Adjust the size as needed
unsigned long previousMillis = 0; // Stores last time the task was run
const long interval = 5000; // 5 seconds

#define FACTORYRESET_ENABLE         1             //BLE Factory Reset 1=RESET
#define MINIMUM_FIRMWARE_VERSION    "0.6.6"       //Declare Min Firmware Version
#define MODE_LED_BEHAVIOUR          "MODE"        //LED activity, valid options are "DISABLE" or "MODE" or "BLEUART" or "HWUART"  or "SPI"  or "MANUAL"

//Objects/Variables
Adafruit_SHT31 sht31 = Adafruit_SHT31();        //Temp & Humidity Object
Adafruit_TSL2591 tsl = Adafruit_TSL2591(2591);  // pass in a number for the sensor identifier (for your use later)
RTC_DS3231 rtc;                                 //RTC Object
Adafruit_BluefruitLE_SPI ble(BLUEFRUIT_SPI_CS, BLUEFRUIT_SPI_IRQ, BLUEFRUIT_SPI_RST);   // Create the bluefruit object, hardware SPI, using SCK/MOSI/MISO hardware SPI pins and then user selected CS/IRQ/RST
Sd2Card card;                                   //SD variable
SdVolume volume;                                //SD variable
SdFile root;                                    //SD variable
                                

//Additional Pre-Setup
void error(const __FlashStringHelper*err) // A small helper for BLE
{
  Serial.println(err);
  while (1);
}

//Initilize
void setup()
{
  //Pin Setup
  pinMode(0, OUTPUT);   // Set RX LED as an output
  pinMode(1, OUTPUT);   // Set RX LED as an output
  pinMode(2, OUTPUT);   // Set RX LED as an output
  pinMode(3, OUTPUT);   // Set RX LED as an output

  //Serial Setup
    Serial.begin(9600);     //This pipes to the serial monitor
    SerialSetupTandH();     //Startup delay to connect Serial Monitor, Check for Temp & Humidity Sensor conneciton
    SerialSetup_TSL2591();  //Startup to connect Serial Monitor, Check for Light & IR Sensor conneciton
    SerialSetup_RTC();
    SD_Setup();             //see if the card is present and can be initialized, Create or open the CSV file
    BLEsetup();             //Startup for BLE Module
}

void loop(void)
{ 
  //Run Once at startup after 100ms Delay
  if (OnePassFlag == 0) {
    delay(100);                                                   // Delay 100msec
    Serial.println(F("Loop Started, Lets Go"));   // Display on LCD display 
    Serial.println(F("------------------------------------"));    //Print Break Line 
    OnePassFlag++;                                                // Increment OnePassFlag
  } 
  //Variables & Function for Date & Time
  DateTime now = rtc.now();           //Setup Time variable
  getTimeAndDateAsCharArray(now, dateTimeString, sizeof(dateTimeString));     // Call the function to format the date and time into the char array
  
  //Variables & Convertions for Temp & Humidity
  float tc = sht31.readTemperature();   //Temp in degC 
  float h = sht31.readHumidity();       //Humidity %
  float tf = (tc*1.8)+32;               //Temp in degF
  
  //Variables & Convertions for Lux Reading
  uint32_t lum = tsl.getFullLuminosity();   //Variable for lumosity
  uint16_t ir, full;                        //Variable definations
  ir = lum >> 16;
  full = lum & 0xFFFF;
  float lux = (tsl.calculateLux(full, ir)); //Variable for lux
  
  // Convert the float to a string
  char *char_dateTimeString = dateTimeString; //Convert char[30] to char*

  // Read and send CSV data
  //sendCSVData();
  
  unsigned long currentMillis = millis();
  if(currentMillis - previousMillis >= interval) //Loop approx every 1 second
  {
    //Save Sensor values to SD card
    SaveCSVtoSD(char_dateTimeString,tf,h,lux);
    Serial.println();

    //Setup and publish values over BLE
    static char dataString[100];  // Buffer to hold the concatenated string (static to retain scope after function ends)
    char char_tf[10];             //Char variable for TempF
    char char_h[10];              //Char variable for Humidity
    char char_lux[10];            //Char variable for Lux

    // Convert the float values to strings with 2 decimal places
    dtostrf(tf, 5, 2, char_tf);                 //Convert Float TempF to Char
    dtostrf(h, 5, 2, char_h);                   //Convert Float Humidity to Char
    dtostrf(lux, 5, 2, char_lux);               //Convert Float Lux to Char
    
    // Initialize the buffer to an empty string
    dataString[0] = '\0';
    //Send to BLE
    snprintf(dataString, sizeof(dataString), "%s|%s|%s", char_tf, char_h,char_lux);
    Serial.print("Sending: ");
    Serial.println(dataString);
    ble.print("AT+BLEUARTTX="); // Send over BLE UART
    ble.println(dataString);
    previousMillis = currentMillis; // Update last run time
  }

}

/*------------------------------------------------------------------------------------------------------------------------------------*/

char* dtostrf(float val, signed char width, unsigned char prec, char *sout)     // Custom dtostrf function for SAMD21
{
  char fmt[20];
  sprintf(fmt, "%%%d.%df", width, prec);
  sprintf(sout, fmt, val);
  return sout;
}
void configureSensor(void)   //TSL2591 Config
{
  // You can change the gain on the fly, to adapt to brighter/dimmer light situations
  //tsl.setGain(TSL2591_GAIN_LOW);    // 1x gain (bright light)
  tsl.setGain(TSL2591_GAIN_MED);      // 25x gain
  //tsl.setGain(TSL2591_GAIN_HIGH);   // 428x gain
  
  // Changing the integration time gives you a longer time over which to sense light
  // longer timelines are slower, but are good in very low light situtations!
  //tsl.setTiming(TSL2591_INTEGRATIONTIME_100MS);  // shortest integration time (bright light)
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_200MS);
  tsl.setTiming(TSL2591_INTEGRATIONTIME_300MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_400MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_500MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_600MS);  // longest integration time (dim light)

  /* Display the gain and integration time for reference sake */  
  //Serial.println(F("------------------------------------"));
  Serial.print  (F("Gain:         "));
  tsl2591Gain_t gain = tsl.getGain();
  switch(gain)
  {
    case TSL2591_GAIN_LOW:
      Serial.println(F("1x (Low)"));
      break;
    case TSL2591_GAIN_MED:
      Serial.println(F("25x (Medium)"));
      break;
    case TSL2591_GAIN_HIGH:
      Serial.println(F("428x (High)"));
      break;
    case TSL2591_GAIN_MAX:
      Serial.println(F("9876x (Max)"));
      break;
  }
  Serial.print  (F("Timing:       "));
  Serial.print((tsl.getTiming() + 1) * 100, DEC); 
  Serial.println(F(" ms"));
  Serial.println(F("------------------------------------"));
  //Serial.println(F(""));
}
void SerialSetup_RTC()      //Startup delay to connect Serial Monitor, Check for RTC conneciton
{
  if (! rtc.begin()) {
     Serial.println("Couldn't find RTC");
      Serial.flush();
      while (1) delay(10);
   }

  if (rtc.lostPower()) {
    Serial.println("RTC lost power, let's set the time!");
    // When time needs to be set on a new device, or after a power loss, the
    // following line sets the RTC to the date & time this sketch was compiled
    rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
  }
}
void SerialSetupTandH()     //Startup delay to connect Serial Monitor, Check for Temp & Humidity Sensor conneciton
{
  while (!Serial)
    //delay(10);     // will pause Zero, Leonardo, etc until serial console opens

  Serial.println(F("SHT31 test"));
  if (! sht31.begin(0x44)) {   // Set to 0x45 for alternate i2c addr
    Serial.println(F("Couldn't find SHT31"));
    while (1) delay(1);
  }
  Serial.print(F("Heater Enabled State: "));
  if (sht31.isHeaterEnabled())
    Serial.println(F("ENABLED"));
  else
    Serial.println(F("DISABLED"));

  Serial.println(F("------------------------------------"));
}
void SerialSetup_TSL2591()  //Startup delay to connect Serial Monitor, Check for Light Sensor conneciton
{
  Serial.println(F("Starting Adafruit TSL2591 Test!"));
  
  if (tsl.begin()) 
  {
    Serial.println(F("Found a TSL2591 sensor"));
  } 
  else 
  {
    Serial.println(F("No sensor found ... check your wiring?"));
    while (1);
  }
  Serial.println(F("------------------------------------"));
  /* Configure the sensor */
  configureSensor();
  // Now we're ready to get readings
}
void getTimeAndDateAsCharArray(DateTime now, char* buffer, size_t bufferSize)      // Format the date and time into the char array
{
  // Format the date and time into the char array
  snprintf(buffer, bufferSize, "%04d/%02d/%02d (%s) %02d:%02d:%02d", 
           now.year(), 
           now.month(), 
           now.day(), 
           daysOfTheWeek[now.dayOfTheWeek()], 
           now.hour(), 
           now.minute(), 
           now.second());
}
void SD_Setup()            //see if the card is present and can be initialized, Create or open the CSV file
{
  // see if the card is present and can be initialized:
  Serial.print("Initializing SD card...");
  if (!SD.begin(3)) {
    Serial.println(F("Card failed, or not present"));
    // don't do anything more:
    return;
  }
  Serial.println(F("card initialized."));

  // Create or open the CSV file
  File dataFile = SD.open("data.csv", FILE_WRITE);
  if(dataFile){
  dataFile.close();
  Serial.println(F("data.csv Created or Existing on SD card. File Closed"));
  }
  else{
  Serial.println(F("Error creating or opening data.csv on SD card!"));
  }
  Serial.println(F("------------------------------------"));
}
void saveData(char* dataString)           //Check for SD card and save dataString to data.csv
{ 
  File sensorData;
  //File sensorData;
  if(SD.exists("data.csv")){ // check the card is still there
  // now append new data file
  sensorData = SD.open("data.csv", FILE_WRITE);
  if (sensorData){
  sensorData.println(dataString);
  sensorData.close(); // close the file
  Serial.println("Data written to data.csv on SD card.");
  }
  }
  else{
  Serial.println("Error writing to file !");
  }
}
void SaveCSVtoSD(const char* DateTimeString, float tf, float h, float lux)
{
  static char dataString[100];  // Buffer to hold the concatenated string (static to retain scope after function ends)
  static char dataString2[100];  // Buffer to hold the concatenated string (static to retain scope after function ends)
  char char_tf[10];             //Char variable for TempF
  char char_h[10];              //Char variable for Humidity
  char char_lux[10];            //Char variable for Lux

  // Convert the float values to strings with 2 decimal places
  dtostrf(tf, 5, 2, char_tf);                 //Convert Float TempF to Char
  dtostrf(h, 5, 2, char_h);                   //Convert Float Humidity to Char
  dtostrf(lux, 5, 2, char_lux);               //Convert Float Lux to Char
  
  // Initialize the buffer to an empty string
  dataString[0] = '\0';

  // Build the CSV string
  snprintf(dataString, sizeof(dataString), "%s,%s,%s,%s", DateTimeString, char_tf, char_h,char_lux);
  // Save to SD card
  saveData(dataString); 
  // Print to Serial Monitor
  Serial.println(dataString);
}
void sendCSVData()            // Read and send CSV data via Bluetooth 
{
  File dataFile = SD.open("data.csv");

  if (dataFile) {
    while (dataFile.available()) {
      // Read a line from the CSV file
      String dataLine = dataFile.readStringUntil('\n');
      
      // Send the line over Bluetooth
      ble.println(dataLine);
      
      // Print the data for debugging
      //Serial.println(dataLine);
    }
    dataFile.close();
    Serial.println("Finished sending CSV data.");
  } else {
    Serial.println("Error opening data.csv");
  }
}
void BLEsetup()             //Sets up the HW an the BLE module (this function is called automatically on startup)
{
  //while (!Serial);  // required for Flora & Micro
  //delay(500);

  Serial.begin(9600);
  Serial.println(F("Adafruit Bluefruit Command Mode Example"));
  Serial.println(F("---------------------------------------"));

  /* Initialise the module */
  Serial.print(F("Initialising the Bluefruit LE module: "));

  if ( !ble.begin(VERBOSE_MODE) )
  {
    error(F("Couldn't find Bluefruit, make sure it's in CoMmanD mode & check wiring?"));
  }
  Serial.println( F("OK!") );

  if ( FACTORYRESET_ENABLE )
  {
    /* Perform a factory reset to make sure everything is in a known state */
    Serial.println(F("Performing a factory reset: "));
    if ( ! ble.factoryReset() ){
      error(F("Couldn't factory reset"));
    }
  }

  // Start advertising
  ble.setMode(BLUEFRUIT_MODE_DATA);
  ble.println("AT+GAPDEVNAME=GardenBoard");
  if (! ble.sendCommandCheckOK("AT+GAPDEVNAME=GardenBoard")) {
    Serial.println("Couldn't set device name");
  }
  ble.sendCommandCheckOK("AT+GAPCONNECTABLE=1");

  /* Disable command echo from Bluefruit */
  ble.echo(false);

  Serial.println("Requesting Bluefruit info:");
  /* Print Bluefruit information */
  ble.info();

  /*
  Serial.println(F("Please use Adafruit Bluefruit LE app to connect in UART mode"));
  Serial.println(F("Then Enter characters to send to Bluefruit"));
  Serial.println();
  */

  ble.verbose(false);  // debug info is a little annoying after this point!

  /* Wait for connection */
  while (! ble.isConnected()) {
      delay(500);
      Serial.println("Waiting on Conneciton");
  }

  // LED Activity command is only supported from 0.6.6
  if ( ble.isVersionAtLeast(MINIMUM_FIRMWARE_VERSION) )
  {
    // Change Mode LED Activity
    Serial.println(F("******************************"));
    Serial.println(F("Change LED activity to " MODE_LED_BEHAVIOUR));
    ble.sendCommandCheckOK("AT+HWModeLED=" MODE_LED_BEHAVIOUR);
    Serial.println(F("******************************"));
  }
}